{
open Parser

let inc_lineno lexbuf =
  let pos = lexbuf.Lexing.lex_curr_p in
    lexbuf.Lexing.lex_curr_p <- { pos with
                                    Lexing.pos_lnum = pos.Lexing.pos_lnum + 1;
                                    Lexing.pos_bol = pos.Lexing.pos_cnum;
                                }

let cur_token = ref None

let return_token tok = cur_token := Some tok; tok

let string_of_token = function
    CLASS      -> "CLASS"
  | EXTENDS    -> "EXTENDS"
  | IMPLEMENTS -> "IMPLEMENTS"
  | EXTENSION  -> "EXTENSION"
  | INTERFACE  -> "INTERFACE"
  | IMPLEMENTATION -> "IMPLEMENTATION"
  | NEW        -> "NEW"
  | NULL       -> "NULL"
  | IF         -> "IF"
  | ELSE       -> "ELSE"
  | TRUE       -> "TRUE"
  | FALSE      -> "FALSE"
  | WHILE      -> "WHILE"
  | EXISTS     -> "EXISTS"
  | WHERE      -> "WHERE"
  | RETURN     -> "RETURN"
  | OF         -> "OF" 
  | STATIC     -> "STATIC"
  | ASSIGN     -> "ASSIGN"
  | EQ         -> "EQ"
  | NEQ        -> "NEQ"
  | LEQ        -> "LEQ"
  | GEQ        -> "GEQ"
  | AND        -> "AND"
  | OR         -> "OR"
  | NOT        -> "NOT"
  | PLUS       -> "PLUS"
  | MINUS      -> "MINUS"
  | TIMES      -> "TIMES"
  | DIV        -> "DIV"
  | COMMA      -> "COMMA"
  | SEMI       -> "SEMI"
  | DOT        -> "DOT"
  | AT         -> "AT"
  | LBRACE     -> "LBRACE"
  | RBRACE     -> "RBRACE"
  | LSQBRACKET -> "LSQBRACKET"
  | RSQBRACKET -> "RSQBRACKET"
  | LPAREN     -> "LPAREN"
  | RPAREN     -> "RPAREN"
  | LABRACKET  -> "LABRACKET"
  | RABRACKET  -> "RABRACKET"
  | ID(x)      -> "ID(" ^ x ^ ")"
  | INT(i)     -> "INT(" ^ string_of_int i ^ ")"
  | EOF        -> "EOF"
  | CAST       -> "CAST"

let error_msg lexbuf = 
  let string_of_position p = 
    p.Lexing.pos_fname ^ ":" ^ string_of_int p.Lexing.pos_lnum ^ ":" 
    ^ string_of_int (p.Lexing.pos_cnum - p.Lexing.pos_bol)
  in
    string_of_position lexbuf.Lexing.lex_curr_p ^ ": parse error" ^
      (match !cur_token with
           None -> ""
         | Some tok -> " (next token: " ^ string_of_token tok ^ ")")

let maybe_keyword id = match id with
    "class" -> CLASS
  | "extends" -> EXTENDS
  | "extension" -> EXTENSION
  | "interface" -> INTERFACE
  | "implementation" -> IMPLEMENTATION
  | "implements" -> IMPLEMENTS
  | "static" -> STATIC
  | "new" -> NEW
  | "null" -> NULL
  | "if" -> IF
  | "else" -> ELSE
  | "true" -> TRUE
  | "false" -> FALSE
  | "exists" -> EXISTS
  | "where" -> WHERE
  | "while" -> WHILE
  | "return" -> RETURN
  | "of" -> OF
  | s -> ID s
}

rule token = parse
    [' ' '\t' '\r']+       { token lexbuf }
  | ['\n']                 { inc_lineno lexbuf; token lexbuf }
  | '='                    { return_token ASSIGN }
  | '=' '='                { return_token EQ }
  | '!' '='                { return_token NEQ }
  | '<'                    { return_token LABRACKET }
  | '>'                    { return_token RABRACKET }
  | '<' '='                { return_token LEQ }
  | '>' '='                { return_token GEQ }
  | '&' '&'                { return_token AND }
  | '|' '|'                { return_token OR }
  | '!'                    { return_token NOT }
  | '+'                    { return_token PLUS }
  | '-'                    { return_token MINUS }
  | '*'                    { return_token TIMES }
  | '/'                    { return_token DIV }
  | ','                    { return_token COMMA }
  | ';'                    { return_token SEMI }
  | '.'                    { return_token DOT }
  | '@'                    { return_token AT }
  | '{'                    { return_token LBRACE }
  | '}'                    { return_token RBRACE }
  | '['                    { return_token LSQBRACKET }
  | ']'                    { return_token RSQBRACKET }
  | '('                    { return_token LPAREN }
  | ')'                    { return_token RPAREN }
  | ['a'-'z' 'A'-'Z' '_'] ['a'-'z' 'A'-'Z' '_' '0'-'9']* as id
                           { return_token (maybe_keyword id) }
  | ['0'-'9']+ as lxm      { return_token (INT (int_of_string lxm)) }
  | eof                    { return_token EOF }
